const TEST_PORT = 8642;
const APP_PORT_START = 33300;
const APP_PORT_COUNT = 700;

const http = require('http');
const child_process = require('child_process');
const crypto = require('crypto');

const pkg = require('./package.json');

const queue = [];
let running = false;
let nextport = 0;

function sendRequest(port, rd) {
	const req = http.request({
		host: '127.0.0.1',
		port,
		method: rd.method,
		path: rd.url,
		headers: rd.headers,
	}, res => {
		if (rd.ended) return;
		rd.response.statusCode = res.statusCode;
		for (let i = 0; i < res.rawHeaders.length; i += 2) {
			rd.response.setHeader(res.rawHeaders[i], res.rawHeaders[i + 1]);
		}
		res.on('data', chunk => rd.response.write(chunk));
		res.on('end', () => {
			if (rd.ended) return;
			rd.ended = true;
			rd.response.end();
		});
	});
	req.on('error', e => {
		console.log(e);
	});
	for (let chunk of rd.body) req.write(chunk);
	req.end();
}

function runQueue() {
	
	if (running || !queue.length) return;
	
	const rd = queue.pop();
	console.log(rd.method, rd.url);
	
	const port = nextport++ % APP_PORT_COUNT + APP_PORT_START;
	
	const bytes = crypto.randomBytes(16).toString('hex');
	const readystr = '=== 2038 ' + bytes + ' READY ===';
	
	const rt = Date.now();
	const app = child_process.spawn(process.argv[0], [pkg[2038].app], {
		env: Object.assign({}, process.env, {
			"_2038_PORT": '' + port,
			"_2038_READY": readystr,
		}),
		stdio: [null, 'pipe', 'inherit'],
	});
	
	let last = '';
	let ready = false;
	app.stdout.on('data', chunk => {
		
		const cstr = chunk.toString('utf8');
		const lines = (last + cstr).split('\n');
		for (let i = 0; i < lines.length; i++) {
			if (!ready && lines[i].trim() === readystr) {
				ready = true;
				sendRequest(port, rd);
			}
			else if (i < lines.length - 1) {
				console.log(lines[i]);
			}
			else {
				last = lines[i];
			}
		}
		
	});
	
	let t = setTimeout(() => {
		console.error('App did not exit within 5 seconds');
		app.kill(15);
		t = setTimeout(() => app.kill(9), 1000);
	}, 5000);
	app.on('exit', code => {
		if (t) clearTimeout(t);
		if (code) console.error('App exited with code', code);
		else console.log('Request', rd.url, 'handled in', Date.now() - rt, 'ms');
		if (!rd.ended) {
			rd.ended = true;
			rd.response.statusCode = 500;
			rd.response.end('App exited with code ' + code);
		}
		running = false;
		runQueue();
	});
	
}

const server = http.createServer(function (req, res) {
	
	const url = req.url;
	const method = req.method;
	
	if (url.indexOf('/!/' + pkg.name) < 0) {
		res.statusCode = 404;
		res.end('Invalid app url');
	}
	
	let body = [];
	req.on('data', chunk => body.push(chunk));
	req.on('end', () => {
		
		queue.push({
			url,
			method,
			headers: req.headers,
			body,
			response: res,
			ended: false,
		});
		
		runQueue();
		
	});
	
});
server.listen(TEST_PORT);

console.log('Go to http://localhost:' + TEST_PORT + '/!/' + pkg.name + ' to test your app');
